@@ -8,8 +8,9 @@ import traceback
 from buildbot import master
 
 class ConfigLoader(master.BuildMaster):
-    def __init__(self, configFileName="master.cfg"):
-        master.BuildMaster.__init__(self, ".", configFileName)
+    def __init__(self, basedir=os.getcwd(), configFileName="master.cfg"):
+        master.BuildMaster.__init__(self, basedir, configFileName)
+        configFileName = os.path.join(basedir, configFileName)
         dir = os.getcwd()
         # Use a temporary directory since loadConfig() creates a bunch of
         # directories and compiles .py files
@@ -18,7 +19,7 @@ class ConfigLoader(master.BuildMaster):
             copy(configFileName, tempdir)
             for entry in os.listdir("."):
                 # Any code in a subdirectory will _not_ be copied! This is a bug
-                if isfile(entry):
+                if isfile(entry) and not entry.startswith("twistd.log"):
                     copy(entry, tempdir)
         except:
             raise
@@ -31,22 +32,8 @@ class ConfigLoader(master.BuildMaster):
             self.loadConfig(configFile)
         except:
             os.chdir(dir)
+            configFile.close()
             rmtree(tempdir)
             raise
         os.chdir(dir)
         rmtree(tempdir)
-
-if __name__ == '__main__':
-    try:
-        if len(sys.argv) > 1:
-            c = ConfigLoader(sys.argv[1])
-        else:
-            c = ConfigLoader()
-    except IOError:
-        print >> sys.stderr, "Could not open config file"
-        sys.exit(2)
-    except:
-        print >> sys.stderr, "Error in config file:"
-        t, v, tb = sys.exc_info()
-        print >> sys.stderr, traceback.print_exception(t, v, tb)
-        sys.exit(1)
