@@ -6,6 +6,8 @@ import os, sys, stat, re, time
 import traceback
 from twisted.python import usage, util, runtime
 
+from buildbot.interfaces import BuildbotNotRunningError
+
 # this is mostly just a front-end for mktap, twistd, and kill(1), but in the
 # future it will also provide an interface to some developer tools that talk
 # directly to a remote buildmaster (like 'try' and a status client)
@@ -317,6 +319,10 @@ class MasterOptions(MakerBase):
         ]
     optParameters = [
         ["config", "c", "master.cfg", "name of the buildmaster config file"],
+        ["log-size", "s", "1000000",
+         "size at which to rotate twisted log files"],
+        ["log-count", "l", "None",
+         "limit the number of kept old twisted log files"],
         ]
     def getSynopsis(self):
         return "Usage:    buildbot create-master [options] <basedir>"
@@ -332,14 +338,35 @@ class MasterOptions(MakerBase):
     See doc/config.xhtml for details about what can be controlled through
     this interface."""
 
+    def postOptions(self):
+        MakerBase.postOptions(self)
+        if not re.match('^\d+$', self['log-size']):
+            raise usage.UsageError("log-size parameter needs to be an int")
+        if not re.match('^\d+$', self['log-count']) and \
+                self['log-count'] != 'None':
+            raise usage.UsageError("log-count parameter needs to be an int "+
+                                   " or None")
+
+
 masterTAC = """
 from twisted.application import service
 from buildbot.master import BuildMaster
 
 basedir = r'%(basedir)s'
 configfile = r'%(config)s'
+rotateLength = %(log-size)s
+maxRotatedFiles = %(log-count)s
 
 application = service.Application('buildmaster')
+try:
+  from twisted.python.logfile import LogFile
+  from twisted.python.log import ILogObserver, FileLogObserver
+  logfile = LogFile.fromFullPath("twistd.log", rotateLength=rotateLength,
+                                 maxRotatedFiles=maxRotatedFiles)
+  application.setComponent(ILogObserver, FileLogObserver(logfile).emit)
+except ImportError:
+  # probably not yet twisted 8.2.0 and beyond, can't set log yet
+  pass
 BuildMaster(basedir, configfile).setServiceParent(application)
 
 """
@@ -372,10 +399,16 @@ class SlaveOptions(MakerBase):
 
         ["keepalive", "k", 600,
          "Interval at which keepalives should be sent (in seconds)"],
-        ["usepty", None, 1,
-         "(1 or 0) child processes should be run in a pty"],
+        ["usepty", None, 0,
+         "(1 or 0) child processes should be run in a pty (default 0)"],
         ["umask", None, "None",
          "controls permissions of generated files. Use --umask=022 to be world-readable"],
+        ["maxdelay", None, 300,
+         "Maximum time between connection attempts"],
+        ["log-size", "s", "1000000",
+         "size at which to rotate twisted log files"],
+        ["log-count", "l", "None",
+         "limit the number of kept old twisted log files"],
         ]
     
     longdesc = """
@@ -396,6 +429,8 @@ class SlaveOptions(MakerBase):
         if len(args) < 4:
             raise usage.UsageError("command needs more arguments")
         basedir, master, name, passwd = args
+        if master[:5] == "http:":
+            raise usage.UsageError("<master> is not a URL - do not use URL")
         self['basedir'] = basedir
         self['master'] = master
         self['name'] = name
@@ -405,8 +440,15 @@ class SlaveOptions(MakerBase):
         MakerBase.postOptions(self)
         self['usepty'] = int(self['usepty'])
         self['keepalive'] = int(self['keepalive'])
+        self['maxdelay'] = int(self['maxdelay'])
         if self['master'].find(":") == -1:
             raise usage.UsageError("--master must be in the form host:portnum")
+        if not re.match('^\d+$', self['log-size']):
+            raise usage.UsageError("log-size parameter needs to be an int")
+        if not re.match('^\d+$', self['log-count']) and \
+                self['log-count'] != 'None':
+            raise usage.UsageError("log-count parameter needs to be an int "+
+                                   " or None")
 
 slaveTAC = """
 from twisted.application import service
@@ -420,10 +462,22 @@ passwd = '%(passwd)s'
 keepalive = %(keepalive)d
 usepty = %(usepty)d
 umask = %(umask)s
+maxdelay = %(maxdelay)d
+rotateLength = %(log-size)s
+maxRotatedFiles = %(log-count)s
 
 application = service.Application('buildslave')
+try:
+  from twisted.python.logfile import LogFile
+  from twisted.python.log import ILogObserver, FileLogObserver
+  logfile = LogFile.fromFullPath("twistd.log", rotateLength=rotateLength,
+                                 maxRotatedFiles=maxRotatedFiles)
+  application.setComponent(ILogObserver, FileLogObserver(logfile).emit)
+except ImportError:
+  # probably not yet twisted 8.2.0 and beyond, can't set log yet
+  pass
 s = BuildSlave(buildmaster_host, port, slavename, passwd, basedir,
-               keepalive, usepty, umask=umask)
+               keepalive, usepty, umask=umask, maxdelay=maxdelay)
 s.setServiceParent(application)
 
 """
@@ -457,7 +511,10 @@ def stop(config, signame="TERM", wait=False):
     basedir = config['basedir']
     quiet = config['quiet']
     os.chdir(basedir)
-    f = open("twistd.pid", "rt")
+    try:
+        f = open("twistd.pid", "rt")
+    except:
+        raise BuildbotNotRunningError
     pid = int(f.read().strip())
     signum = getattr(signal, "SIG"+signame)
     timer = 0
@@ -483,7 +540,10 @@ def stop(config, signame="TERM", wait=False):
 def restart(config):
     quiet = config['quiet']
     from buildbot.scripts.startup import start
-    stop(config, wait=True)
+    try:
+        stop(config, wait=True)
+    except BuildbotNotRunningError:
+        pass
     if not quiet:
         print "now restarting buildbot process.."
     start(config)
@@ -659,12 +719,14 @@ class SendChangeOptions(usage.Options):
          "Location of the buildmaster's PBListener (host:port)"),
         ("username", "u", None, "Username performing the commit"),
         ("branch", "b", None, "Branch specifier"),
+        ("category", "c", None, "Category of repository"),
         ("revision", "r", None, "Revision specifier (string)"),
         ("revision_number", "n", None, "Revision specifier (integer)"),
         ("revision_file", None, None, "Filename containing revision spec"),
         ("comments", "m", None, "log message"),
         ("logfile", "F", None,
          "Read the log messages from this file (- for stdin)"),
+        ("when", "w", None, "timestamp to use as the change time"),
         ]
     def getSynopsis(self):
         return "Usage:    buildbot sendchange [options] filenames.."
@@ -681,7 +743,12 @@ def sendchange(config, runReactor=False):
     user = config.get('username', opts.get('username'))
     master = config.get('master', opts.get('master'))
     branch = config.get('branch', opts.get('branch'))
+    category = config.get('category', opts.get('category'))
     revision = config.get('revision')
+    if config.get('when'):
+        when = float(config.get('when'))
+    else:
+        when = None
     # SVN and P4 use numeric revisions
     if config.get("revision_number"):
         revision = int(config['revision_number'])
@@ -704,7 +771,7 @@ def sendchange(config, runReactor=False):
     assert master, "you must provide the master location"
 
     s = Sender(master, user)
-    d = s.send(branch, revision, comments, files)
+    d = s.send(branch, revision, comments, files, category=category, when=when)
     if runReactor:
         d.addCallbacks(s.printSuccess, s.printFailure)
         d.addBoth(s.stop)
@@ -769,6 +836,7 @@ class TryOptions(usage.Options):
 
     optFlags = [
         ["wait", None, "wait until the builds have finished"],
+        ["dryrun", 'n', "Gather info, but don't actually submit."],
         ]
 
     def __init__(self):
@@ -841,10 +909,13 @@ class CheckConfigOptions(usage.Options):
 
 def doCheckConfig(config):
     quiet = config.get('quiet')
-    configFile = config.get('configFile')
+    configFileName = config.get('configFile')
     try:
         from buildbot.scripts.checkconfig import ConfigLoader
-        ConfigLoader(configFile)
+        if os.path.isdir(configFileName):
+            ConfigLoader(basedir=configFileName)
+        else:
+            ConfigLoader(configFileName=configFileName)
     except:
         if not quiet:
             # Print out the traceback in a nice format
