@@ -36,7 +36,9 @@ class SourceStamp(util.ComparableMixin):
         self.patch = patch
         if changes:
             self.changes = tuple(changes)
-            self.branch = changes[0].branch
+            # set branch and revision to most recent change
+            self.branch = changes[-1].branch
+            self.revision = changes[-1].revision
 
     def canBeMergedWith(self, other):
         if other.branch != self.branch:
@@ -80,6 +82,9 @@ class SourceStamp(util.ComparableMixin):
                                 changes=changes)
         return newsource
 
+    def getAbsoluteSourceStamp(self, got_revision):
+        return SourceStamp(branch=self.branch, revision=got_revision, patch=self.patch)
+
     def getText(self):
         # TODO: this won't work for VC's with huge 'revision' strings
         if self.revision is None:
