@@ -3,12 +3,16 @@
 from zope.interface import implements
 from twisted.python import log
 from twisted.persisted import styles
-from twisted.internet import reactor, defer
+from twisted.internet import reactor, defer, threads
 from twisted.protocols import basic
+from buildbot.process.properties import Properties
 
+import weakref
 import os, shutil, sys, re, urllib, itertools
+import gc
 from cPickle import load, dump
 from cStringIO import StringIO
+from bz2 import BZ2File
 
 # sibling imports
 from buildbot import interfaces, util, sourcestamp
@@ -183,6 +187,22 @@ class LogFileProducer:
             self.consumer.finish()
             self.consumer = None
 
+def _tryremove(filename, timeout, retries):
+    """Try to remove a file, and if failed, try again in timeout.
+    Increases the timeout by a factor of 4, and only keeps trying for
+    another retries-amount of times.
+
+    """
+    try:
+        os.unlink(filename)
+    except OSError:
+        if retries > 0:
+            reactor.callLater(timeout, _tryremove, filename, timeout * 4, 
+                              retries - 1)
+        else:
+            log.msg("giving up on removing %s after over %d seconds" %
+                    (filename, timeout))
+
 class LogFile:
     """A LogFile keeps all of its contents on disk, in a non-pickle format to
     which new entries can easily be appended. The file on disk has a name
@@ -236,7 +256,8 @@ class LogFile:
         return os.path.join(self.step.build.builder.basedir, self.filename)
 
     def hasContents(self):
-        return os.path.exists(self.getFilename())
+        return os.path.exists(self.getFilename() + '.bz2') or \
+            os.path.exists(self.getFilename())
 
     def getName(self):
         return self.name
@@ -260,6 +281,11 @@ class LogFile:
             # don't close it!
             return self.openfile
         # otherwise they get their own read-only handle
+        # try a compressed log first
+        try:
+            return BZ2File(self.getFilename() + ".bz2", "r")
+        except IOError:
+            pass
         return open(self.getFilename(), "r")
 
     def getText(self):
@@ -406,6 +432,7 @@ class LogFile:
             # filehandle will be released and automatically closed. We will
             # do a sync, however, to make sure the log gets saved in case of
             # a crash.
+            self.openfile.flush()
             os.fsync(self.openfile.fileno())
             del self.openfile
         self.finished = True
@@ -415,6 +442,41 @@ class LogFile:
             w.callback(self)
         self.watchers = []
 
+
+    def compressLog(self):
+        compressed = self.getFilename() + ".bz2.tmp"
+        d = threads.deferToThread(self._compressLog, compressed)
+        d.addCallback(self._renameCompressedLog, compressed)
+        d.addErrback(self._cleanupFailedCompress, compressed)
+        return d
+
+    def _compressLog(self, compressed):
+        infile = self.getFile()
+        cf = BZ2File(compressed, 'w')
+        bufsize = 1024*1024
+        while True:
+            buf = infile.read(bufsize)
+            cf.write(buf)
+            if len(buf) < bufsize:
+                break
+        cf.close()
+    def _renameCompressedLog(self, rv, compressed):
+        filename = self.getFilename() + '.bz2'
+        if sys.platform == 'win32':
+            # windows cannot rename a file on top of an existing one, so
+            # fall back to delete-first. There are ways this can fail and
+            # lose the builder's history, so we avoid using it in the
+            # general (non-windows) case
+            if os.path.exists(filename):
+                os.unlink(filename)
+        os.rename(compressed, filename)
+        _tryremove(self.getFilename(), 1, 5)
+    def _cleanupFailedCompress(self, failure, compressed):
+        log.msg("failed to compress %s" % self.getFilename())
+        if os.path.exists(compressed):
+            _tryremove(compressed, 1, 5)
+        failure.trap() # reraise the failure
+
     # persistence stuff
     def __getstate__(self):
         d = self.__dict__.copy()
@@ -448,7 +510,6 @@ class LogFile:
             self.finish() # releases self.openfile, which will be closed
         del self.entries
 
-
 class HTMLLogFile:
     implements(interfaces.IStatusLog)
 
@@ -502,15 +563,12 @@ class Event:
     started = None
     finished = None
     text = []
-    color = None
 
     # IStatusEvent methods
     def getTimes(self):
         return (self.started, self.finished)
     def getText(self):
         return self.text
-    def getColor(self):
-        return self.color
     def getLogs(self):
         return []
 
@@ -614,6 +672,7 @@ class BuildRequestStatus:
         self.builderName = builderName
         self.builds = [] # list of BuildStatus objects
         self.observers = []
+        self.submittedAt = None
 
     def buildStarted(self, build):
         self.builds.append(build)
@@ -635,17 +694,22 @@ class BuildRequestStatus:
     def unsubscribe(self, observer):
         self.observers.remove(observer)
 
+    def getSubmitTime(self):
+        return self.submittedAt
+    def setSubmitTime(self, t):
+        self.submittedAt = t
+
 
 class BuildStepStatus(styles.Versioned):
     """
     I represent a collection of output status for a
     L{buildbot.process.step.BuildStep}.
 
-    @type color: string
-    @cvar color: color that this step feels best represents its
-                 current mood. yellow,green,red,orange are the
-                 most likely choices, although purple indicates
-                 an exception
+    Statistics contain any information gleaned from a step that is
+    not in the form of a logfile.  As an example, steps that run
+    tests might gather statistics about the number of passed, failed,
+    or skipped tests.
+
     @type progress: L{buildbot.status.progress.StepProgress}
     @cvar progress: tracks ETA for the step
     @type text: list of strings
@@ -654,22 +718,24 @@ class BuildStepStatus(styles.Versioned):
     @cvar text2: list of short texts added to the overall build description
     @type logs: dict of string -> L{buildbot.status.builder.LogFile}
     @ivar logs: logs of steps
+    @type statistics: dict
+    @ivar statistics: results from running this step
     """
     # note that these are created when the Build is set up, before each
     # corresponding BuildStep has started.
     implements(interfaces.IBuildStepStatus, interfaces.IStatusEvent)
-    persistenceVersion = 1
+    persistenceVersion = 2
 
     started = None
     finished = None
     progress = None
     text = []
-    color = None
     results = (None, [])
     text2 = []
     watchers = []
     updates = {}
     finishedWatchers = []
+    statistics = {}
 
     def __init__(self, parent):
         assert interfaces.IBuildStatus(parent)
@@ -679,6 +745,7 @@ class BuildStepStatus(styles.Versioned):
         self.watchers = []
         self.updates = {}
         self.finishedWatchers = []
+        self.statistics = {}
 
     def getName(self):
         """Returns a short string with the name of this step. This string
@@ -709,6 +776,9 @@ class BuildStepStatus(styles.Versioned):
     def getURLs(self):
         return self.urls.copy()
 
+    def isStarted(self):
+	return (self.started is not None)
+
     def isFinished(self):
         return (self.finished is not None)
 
@@ -742,12 +812,6 @@ class BuildStepStatus(styles.Versioned):
         presenting them to the user."""
         return self.text
 
-    def getColor(self):
-        """Returns a single string with the color that should be used to
-        display this step. 'green', 'orange', 'red', 'yellow' and 'purple'
-        are the most likely ones."""
-        return self.color
-
     def getResults(self):
         """Return a tuple describing the results of the step.
         'result' is one of the constants in L{buildbot.status.builder}:
@@ -763,6 +827,16 @@ class BuildStepStatus(styles.Versioned):
         """
         return (self.results, self.text2)
 
+    def hasStatistic(self, name):
+        """Return true if this step has a value for the given statistic.
+        """
+        return self.statistics.has_key(name)
+
+    def getStatistic(self, name, default=None):
+        """Return the given statistic, if present
+        """
+        return self.statistics.get(name, default)
+
     # subscription interface
 
     def subscribe(self, receiver, updateInterval=10):
@@ -796,6 +870,9 @@ class BuildStepStatus(styles.Versioned):
     def setName(self, stepname):
         self.name = stepname
 
+    def setColor(self, color):
+        log.msg("BuildStepStatus.setColor is no longer supported -- ignoring color %s" % (color,))
+
     def setProgress(self, stepprogress):
         self.progress = stepprogress
 
@@ -839,19 +916,34 @@ class BuildStepStatus(styles.Versioned):
     def addURL(self, name, url):
         self.urls[name] = url
 
-    def setColor(self, color):
-        self.color = color
     def setText(self, text):
         self.text = text
+        for w in self.watchers:
+            w.stepTextChanged(self.build, self, text)
     def setText2(self, text):
         self.text2 = text
+        for w in self.watchers:
+            w.stepText2Changed(self.build, self, text)
+
+    def setStatistic(self, name, value):
+        """Set the given statistic.  Usually called by subclasses.
+        """
+        self.statistics[name] = value
 
     def stepFinished(self, results):
         self.finished = util.now()
         self.results = results
+        cld = [] # deferreds for log compression
+        logCompressionLimit = self.build.builder.logCompressionLimit
         for loog in self.logs:
             if not loog.isFinished():
                 loog.finish()
+            # if log compression is on, and it's a real LogFile,
+            # HTMLLogFiles aren't files
+            if logCompressionLimit is not False and \
+                    isinstance(loog, LogFile):
+                if os.path.getsize(loog.getFilename()) > logCompressionLimit:
+                    cld.append(loog.compressLog())
 
         for r in self.updates.keys():
             if self.updates[r] is not None:
@@ -862,6 +954,12 @@ class BuildStepStatus(styles.Versioned):
         self.finishedWatchers = []
         for w in watchers:
             w.callback(self)
+        if cld:
+            return defer.DeferredList(cld)
+
+    def checkLogfiles(self):
+        # filter out logs that have been deleted
+        self.logs = [ l for l in self.logs if l.hasContents() ]
 
     # persistence
 
@@ -878,6 +976,8 @@ class BuildStepStatus(styles.Versioned):
     def __setstate__(self, d):
         styles.Versioned.__setstate__(self, d)
         # self.build must be filled in by our parent
+
+        # point the logs to this object
         for loog in self.logs:
             loog.step = self
 
@@ -885,21 +985,25 @@ class BuildStepStatus(styles.Versioned):
         if not hasattr(self, "urls"):
             self.urls = {}
 
+    def upgradeToVersion2(self):
+        if not hasattr(self, "statistics"):
+            self.statistics = {}
+
 
 class BuildStatus(styles.Versioned):
     implements(interfaces.IBuildStatus, interfaces.IStatusEvent)
-    persistenceVersion = 2
+    persistenceVersion = 3
 
     source = None
     reason = None
     changes = []
     blamelist = []
+    requests = []
     progress = None
     started = None
     finished = None
     currentStep = None
     text = []
-    color = None
     results = None
     slavename = "???"
 
@@ -924,7 +1028,11 @@ class BuildStatus(styles.Versioned):
         self.finishedWatchers = []
         self.steps = []
         self.testResults = {}
-        self.properties = {}
+        self.properties = Properties()
+        self.requests = []
+
+    def __repr__(self):
+        return "<%s #%s>" % (self.__class__.__name__, self.number)
 
     # IBuildStatus
 
@@ -937,6 +1045,9 @@ class BuildStatus(styles.Versioned):
     def getProperty(self, propname):
         return self.properties[propname]
 
+    def getProperties(self):
+        return self.properties
+
     def getNumber(self):
         return self.number
 
@@ -945,8 +1056,10 @@ class BuildStatus(styles.Versioned):
             return None
         return self.builder.getBuild(self.number-1)
 
-    def getSourceStamp(self):
-        return self.source
+    def getSourceStamp(self, absolute=False):
+        if not absolute or not self.properties.has_key('got_revision'):
+            return self.source
+        return self.source.getAbsoluteSourceStamp(self.properties['got_revision'])
 
     def getReason(self):
         return self.reason
@@ -954,12 +1067,15 @@ class BuildStatus(styles.Versioned):
     def getChanges(self):
         return self.changes
 
+    def getRequests(self):
+        return self.requests
+
     def getResponsibleUsers(self):
         return self.blamelist
 
     def getInterestedUsers(self):
         # TODO: the Builder should add others: sheriffs, domain-owners
-        return self.blamelist
+        return self.blamelist + self.properties.getProperty('owners', [])
 
     def getSteps(self):
         """Return a list of IBuildStepStatus objects. For invariant builds
@@ -972,6 +1088,23 @@ class BuildStatus(styles.Versioned):
     def getTimes(self):
         return (self.started, self.finished)
 
+    _sentinel = [] # used as a sentinel to indicate unspecified initial_value
+    def getSummaryStatistic(self, name, summary_fn, initial_value=_sentinel):
+        """Summarize the named statistic over all steps in which it
+        exists, using combination_fn and initial_value to combine multiple
+        results into a single result.  This translates to a call to Python's
+        X{reduce}::
+            return reduce(summary_fn, step_stats_list, initial_value)
+        """
+        step_stats_list = [
+                st.getStatistic(name)
+                for st in self.steps
+                if st.hasStatistic(name) ]
+        if initial_value is self._sentinel:
+            return reduce(summary_fn, step_stats_list)
+        else:
+            return reduce(summary_fn, step_stats_list, initial_value)
+
     def isFinished(self):
         return (self.finished is not None)
 
@@ -1009,9 +1142,6 @@ class BuildStatus(styles.Versioned):
             text.extend(s.text2)
         return text
 
-    def getColor(self):
-        return self.color
-
     def getResults(self):
         return self.results
 
@@ -1072,8 +1202,8 @@ class BuildStatus(styles.Versioned):
         self.steps.append(s)
         return s
 
-    def setProperty(self, propname, value):
-        self.properties[propname] = value
+    def setProperty(self, propname, value, source):
+        self.properties.setProperty(propname, value, source)
 
     def addTestResult(self, result):
         self.testResults[result.getName()] = result
@@ -1082,6 +1212,9 @@ class BuildStatus(styles.Versioned):
         self.source = sourceStamp
         self.changes = self.source.changes
 
+    def setRequests(self, requests):
+        self.requests = requests
+
     def setReason(self, reason):
         self.reason = reason
     def setBlamelist(self, blamelist):
@@ -1104,8 +1237,6 @@ class BuildStatus(styles.Versioned):
     def setText(self, text):
         assert isinstance(text, (list, tuple))
         self.text = text
-    def setColor(self, color):
-        self.color = color
     def setResults(self, results):
         self.results = results
 
@@ -1147,12 +1278,6 @@ class BuildStatus(styles.Versioned):
 
     # methods called by our BuilderStatus parent
 
-    def pruneLogs(self):
-        # this build is somewhat old: remove the build logs to save space
-        # TODO: delete logs visible through IBuildStatus.getLogs
-        for s in self.steps:
-            s.pruneLogs()
-
     def pruneSteps(self):
         # this build is very old: remove the build steps too
         self.steps = []
@@ -1196,10 +1321,8 @@ class BuildStatus(styles.Versioned):
             # was interrupted. The builder will have a 'shutdown' event, but
             # someone looking at just this build will be confused as to why
             # the last log is truncated.
-        del d['builder'] # filled in by our parent when loading
-        del d['watchers']
-        del d['updates']
-        del d['finishedWatchers']
+        for k in 'builder', 'watchers', 'updates', 'requests', 'finishedWatchers':
+            if k in d: del d[k]
         return d
 
     def __setstate__(self, d):
@@ -1227,6 +1350,12 @@ class BuildStatus(styles.Versioned):
     def upgradeToVersion2(self):
         self.properties = {}
 
+    def upgradeToVersion3(self):
+        # in version 3, self.properties became a Properties object
+        propdict = self.properties
+        self.properties = Properties()
+        self.properties.update(propdict, "Upgrade from previous version")
+
     def upgradeLogfiles(self):
         # upgrade any LogFiles that need it. This must occur after we've been
         # attached to our Builder, and after we know about all LogFiles of
@@ -1242,6 +1371,12 @@ class BuildStatus(styles.Versioned):
                     # transferring its contents onto disk if necessary
                     l.upgrade(logfilename)
 
+    def checkLogfiles(self):
+        # check that all logfiles exist, and remove references to any that
+        # have been deleted (e.g., by purge())
+        for s in self.steps:
+            s.checkLogfiles()
+
     def saveYourself(self):
         filename = os.path.join(self.builder.basedir, "%d" % self.number)
         if os.path.isdir(filename):
@@ -1276,8 +1411,8 @@ class BuilderStatus(styles.Versioned):
     I am responsible for maintaining the list of historic Events and Builds,
     pruning old ones, and loading them from / saving them to disk.
 
-    I live in the buildbot.process.base.Builder object, in the .statusbag
-    attribute.
+    I live in the buildbot.process.base.Builder object, in the
+    .builder_status attribute.
 
     @type  category: string
     @ivar  category: user-defined category this builder belongs to; can be
@@ -1290,9 +1425,12 @@ class BuilderStatus(styles.Versioned):
     # these limit the amount of memory we consume, as well as the size of the
     # main Builder pickle. The Build and LogFile pickles on disk must be
     # handled separately.
-    buildCacheSize = 30
+    buildCacheSize = 15
+    eventHorizon = 50 # forget events beyond this
+
+    # these limit on-disk storage
+    logHorizon = 40 # forget logs in steps in builds beyond this
     buildHorizon = 100 # forget builds beyond this
-    stepHorizon = 50 # forget steps in builds beyond this
 
     category = None
     currentBigState = "offline" # or idle/waiting/interlocked/building
@@ -1313,7 +1451,9 @@ class BuilderStatus(styles.Versioned):
         self.pendingBuilds = []
         self.nextBuild = None
         self.watchers = []
-        self.buildCache = [] # TODO: age builds out of the cache
+        self.buildCache = weakref.WeakValueDictionary()
+        self.buildCache_LRU = []
+        self.logCompressionLimit = False # default to no compression for tests
 
     # persistence
 
@@ -1325,6 +1465,7 @@ class BuilderStatus(styles.Versioned):
         d = styles.Versioned.__getstate__(self)
         d['watchers'] = []
         del d['buildCache']
+        del d['buildCache_LRU']
         for b in self.currentBuilds:
             b.saveYourself()
             # TODO: push a 'hey, build was interrupted' event
@@ -1340,7 +1481,8 @@ class BuilderStatus(styles.Versioned):
         # when loading, re-initialize the transient stuff. Remember that
         # upgradeToVersion1 and such will be called after this finishes.
         styles.Versioned.__setstate__(self, d)
-        self.buildCache = []
+        self.buildCache = weakref.WeakValueDictionary()
+        self.buildCache_LRU = []
         self.currentBuilds = []
         self.pendingBuilds = []
         self.watchers = []
@@ -1348,6 +1490,18 @@ class BuilderStatus(styles.Versioned):
         # self.basedir must be filled in by our parent
         # self.status must be filled in by our parent
 
+    def reconfigFromBuildmaster(self, buildmaster):
+        # Note that we do not hang onto the buildmaster, since this object
+        # gets pickled and unpickled.
+        if buildmaster.buildCacheSize:
+            self.buildCacheSize = buildmaster.buildCacheSize
+        if buildmaster.eventHorizon:
+            self.eventHorizon = buildmaster.eventHorizon
+        if buildmaster.logHorizon:
+            self.logHorizon = buildmaster.logHorizon
+        if buildmaster.buildHorizon:
+            self.buildHorizon = buildmaster.buildHorizon
+
     def upgradeToVersion1(self):
         if hasattr(self, 'slavename'):
             self.slavenames = [self.slavename]
@@ -1369,8 +1523,11 @@ class BuilderStatus(styles.Versioned):
         else:
             self.nextBuildNumber = 0
 
+    def setLogCompressionLimit(self, lowerLimit):
+        self.logCompressionLimit = lowerLimit
+
     def saveYourself(self):
-        for b in self.buildCache:
+        for b in self.currentBuilds:
             if not b.isFinished:
                 # interrupted build, need to save it anyway.
                 # BuildStatus.saveYourself will mark it as interrupted.
@@ -1391,41 +1548,90 @@ class BuilderStatus(styles.Versioned):
 
     # build cache management
 
-    def addBuildToCache(self, build):
-        if build in self.buildCache:
-            return
-        self.buildCache.append(build)
-        while len(self.buildCache) > self.buildCacheSize:
-            self.buildCache.pop(0)
+    def makeBuildFilename(self, number):
+        return os.path.join(self.basedir, "%d" % number)
+
+    def touchBuildCache(self, build):
+        self.buildCache[build.number] = build
+        if build in self.buildCache_LRU:
+            self.buildCache_LRU.remove(build)
+        self.buildCache_LRU = self.buildCache_LRU[-(self.buildCacheSize-1):] + [ build ]
+        return build
 
     def getBuildByNumber(self, number):
+        # first look in currentBuilds
         for b in self.currentBuilds:
             if b.number == number:
-                return b
-        for build in self.buildCache:
-            if build.number == number:
-                return build
-        filename = os.path.join(self.basedir, "%d" % number)
+                return self.touchBuildCache(b)
+
+        # then in the buildCache
+        if number in self.buildCache:
+            return self.touchBuildCache(self.buildCache[number])
+
+        # then fall back to loading it from disk
+        filename = self.makeBuildFilename(number)
         try:
+            log.msg("Loading builder %s's build %d from on-disk pickle"
+                % (self.name, number))
             build = load(open(filename, "rb"))
             styles.doUpgrade()
             build.builder = self
             # handle LogFiles from after 0.5.0 and before 0.6.5
             build.upgradeLogfiles()
-            self.addBuildToCache(build)
-            return build
+            # check that logfiles exist
+            build.checkLogfiles()
+            return self.touchBuildCache(build)
         except IOError:
             raise IndexError("no such build %d" % number)
         except EOFError:
             raise IndexError("corrupted build pickle %d" % number)
 
     def prune(self):
-        return # TODO: change this to walk through the filesystem
-        # first, blow away all builds beyond our build horizon
-        self.builds = self.builds[-self.buildHorizon:]
-        # then prune steps in builds past the step horizon
-        for b in self.builds[0:-self.stepHorizon]:
-            b.pruneSteps()
+        gc.collect()
+
+        # begin by pruning our own events
+        self.events = self.events[-self.eventHorizon:]
+
+        # get the horizons straight
+        if self.buildHorizon:
+            earliest_build = self.nextBuildNumber - self.buildHorizon
+        else:
+            earliest_build = 0
+
+        if self.logHorizon:
+            earliest_log = self.nextBuildNumber - self.logHorizon
+        else:
+            earliest_log = 0
+
+        if earliest_log < earliest_build:
+            earliest_log = earliest_build
+
+        if earliest_build == 0:
+            return
+
+        # skim the directory and delete anything that shouldn't be there anymore
+        build_re = re.compile(r"^([0-9]+)$")
+        build_log_re = re.compile(r"^([0-9]+)-.*$")
+        for filename in os.listdir(self.basedir):
+            num = None
+            mo = build_re.match(filename)
+            is_logfile = False
+            if mo:
+                num = int(mo.group(1))
+            else:
+                mo = build_log_re.match(filename)
+                if mo:
+                    num = int(mo.group(1))
+                    is_logfile = True
+
+            if num is None: continue
+            if num in self.buildCache: continue
+
+            if (is_logfile and num < earliest_log) or num < earliest_build:
+                pathname = os.path.join(self.basedir, filename)
+                log.msg("pruning '%s'" % pathname)
+                try: os.unlink(pathname)
+                except OSError: pass
 
     # IBuilderStatus methods
     def getName(self):
@@ -1449,6 +1655,9 @@ class BuilderStatus(styles.Versioned):
             b = self.getBuild(-2)
         return b
 
+    def getCategory(self):
+        return self.category
+
     def getBuild(self, number):
         if number < 0:
             number = self.nextBuildNumber + number
@@ -1498,7 +1707,7 @@ class BuilderStatus(styles.Versioned):
                 if got >= num_builds:
                     return
 
-    def eventGenerator(self, branches=[]):
+    def eventGenerator(self, branches=[], categories=[]):
         """This function creates a generator which will provide all of this
         Builder's status events, starting with the most recent and
         progressing backwards in time. """
@@ -1520,6 +1729,8 @@ class BuilderStatus(styles.Versioned):
                 break
             if branches and not b.getSourceStamp().branch in branches:
                 continue
+            if categories and not b.getBuilder().getCategory() in categories:
+                continue
             steps = b.getSteps()
             for Ns in range(1, len(steps)+1):
                 if steps[-Ns].started:
@@ -1548,24 +1759,22 @@ class BuilderStatus(styles.Versioned):
     def setSlavenames(self, names):
         self.slavenames = names
 
-    def addEvent(self, text=[], color=None):
+    def addEvent(self, text=[]):
         # this adds a duration event. When it is done, the user should call
-        # e.finish(). They can also mangle it by modifying .text and .color
+        # e.finish(). They can also mangle it by modifying .text
         e = Event()
         e.started = util.now()
         e.text = text
-        e.color = color
         self.events.append(e)
         return e # they are free to mangle it further
 
-    def addPointEvent(self, text=[], color=None):
+    def addPointEvent(self, text=[]):
         # this adds a point event, one which occurs as a single atomic
         # instant of time.
         e = Event()
         e.started = util.now()
         e.finished = 0
         e.text = text
-        e.color = color
         self.events.append(e)
         return e # for consistency, but they really shouldn't touch it
 
@@ -1583,7 +1792,11 @@ class BuilderStatus(styles.Versioned):
             target.builderChangedState(self.name, state)
             return
         for w in self.watchers:
-            w.builderChangedState(self.name, state)
+            try:
+                w.builderChangedState(self.name, state)
+            except:
+                log.msg("Exception caught publishing state to %r" % w)
+                log.err()
 
     def newBuild(self):
         """The Builder has decided to start a build, but the Build object is
@@ -1601,8 +1814,14 @@ class BuilderStatus(styles.Versioned):
 
     def addBuildRequest(self, brstatus):
         self.pendingBuilds.append(brstatus)
-    def removeBuildRequest(self, brstatus):
+        for w in self.watchers:
+            w.requestSubmitted(brstatus)
+
+    def removeBuildRequest(self, brstatus, cancelled=False):
         self.pendingBuilds.remove(brstatus)
+        if cancelled:
+            for w in self.watchers:
+                w.requestCancelled(self, brstatus)
 
     # buildStarted is called by our child BuildStatus instances
     def buildStarted(self, s):
@@ -1613,21 +1832,24 @@ class BuilderStatus(styles.Versioned):
         assert s.number == self.nextBuildNumber - 1
         assert s not in self.currentBuilds
         self.currentBuilds.append(s)
-        self.addBuildToCache(s)
+        self.touchBuildCache(s)
 
         # now that the BuildStatus is prepared to answer queries, we can
         # announce the new build to all our watchers
 
         for w in self.watchers: # TODO: maybe do this later? callLater(0)?
-            receiver = w.buildStarted(self.getName(), s)
-            if receiver:
-                if type(receiver) == type(()):
-                    s.subscribe(receiver[0], receiver[1])
-                else:
-                    s.subscribe(receiver)
-                d = s.waitUntilFinished()
-                d.addCallback(lambda s: s.unsubscribe(receiver))
-
+            try:
+                receiver = w.buildStarted(self.getName(), s)
+                if receiver:
+                    if type(receiver) == type(()):
+                        s.subscribe(receiver[0], receiver[1])
+                    else:
+                        s.subscribe(receiver)
+                    d = s.waitUntilFinished()
+                    d.addCallback(lambda s: s.unsubscribe(receiver))
+            except:
+                log.msg("Exception caught notifying %r of buildStarted event" % w)
+                log.err()
 
     def _buildFinished(self, s):
         assert s in self.currentBuilds
@@ -1637,7 +1859,11 @@ class BuilderStatus(styles.Versioned):
         name = self.getName()
         results = s.getResults()
         for w in self.watchers:
-            w.buildFinished(name, s, results)
+            try:
+                w.buildFinished(name, s, results)
+            except:
+                log.msg("Exception caught notifying %r of buildFinished event" % w)
+                log.err()
 
         self.prune() # conserve disk
 
@@ -1727,10 +1953,13 @@ class SlaveStatus:
     admin = None
     host = None
     connected = False
+    graceful_shutdown = False
 
     def __init__(self, name):
         self.name = name
         self._lastMessageReceived = 0
+        self.runningBuilds = []
+        self.graceful_callbacks = []
 
     def getName(self):
         return self.name
@@ -1742,6 +1971,8 @@ class SlaveStatus:
         return self.connected
     def lastMessageReceived(self):
         return self._lastMessageReceived
+    def getRunningBuilds(self):
+        return self.runningBuilds
 
     def setAdmin(self, admin):
         self.admin = admin
@@ -1752,6 +1983,30 @@ class SlaveStatus:
     def setLastMessageReceived(self, when):
         self._lastMessageReceived = when
 
+    def buildStarted(self, build):
+        self.runningBuilds.append(build)
+    def buildFinished(self, build):
+        self.runningBuilds.remove(build)
+
+    def getGraceful(self):
+        """Return the graceful shutdown flag"""
+        return self.graceful_shutdown
+    def setGraceful(self, graceful):
+        """Set the graceful shutdown flag, and notify all the watchers"""
+        self.graceful_shutdown = graceful
+        for cb in self.graceful_callbacks:
+            reactor.callLater(0, cb, graceful)
+    def addGracefulWatcher(self, watcher):
+        """Add watcher to the list of watchers to be notified when the
+        graceful shutdown flag is changed."""
+        if not watcher in self.graceful_callbacks:
+            self.graceful_callbacks.append(watcher)
+    def removeGracefulWatcher(self, watcher):
+        """Remove watcher from the list of watchers to be notified when the
+        graceful shutdown flag is changed."""
+        if watcher in self.graceful_callbacks:
+            self.graceful_callbacks.remove(watcher)
+
 class Status:
     """
     I represent the status of the buildmaster.
@@ -1777,6 +2032,8 @@ class Status:
         self.watchers = []
         self.activeBuildSets = []
         assert os.path.isdir(basedir)
+        # compress logs bigger than 4k, a good default on linux
+        self.logCompressionLimit = 4*1024
 
 
     # methods called by our clients
@@ -1798,23 +2055,23 @@ class Status:
             pass
         if interfaces.IBuilderStatus.providedBy(thing):
             builder = thing
-            return prefix + urllib.quote(builder.getName(), safe='')
+            return prefix + "builders/%s" % (
+                urllib.quote(builder.getName(), safe=''),
+                )
         if interfaces.IBuildStatus.providedBy(thing):
             build = thing
             builder = build.getBuilder()
-            return "%s%s/builds/%d" % (
-                prefix,
+            return prefix + "builders/%s/builds/%d" % (
                 urllib.quote(builder.getName(), safe=''),
                 build.getNumber())
         if interfaces.IBuildStepStatus.providedBy(thing):
             step = thing
             build = step.getBuild()
             builder = build.getBuilder()
-            return "%s%s/builds/%d/%s" % (
-                prefix,
+            return prefix + "builders/%s/builds/%d/steps/%s" % (
                 urllib.quote(builder.getName(), safe=''),
                 build.getNumber(),
-                "step-" + urllib.quote(step.getName(), safe=''))
+                urllib.quote(step.getName(), safe=''))
         # IBuildSetStatus
         # IBuildRequestStatus
         # ISlaveStatus
@@ -1840,12 +2097,11 @@ class Status:
                     break
             else:
                 return None
-            return "%s%s/builds/%d/%s/%d" % (
-                prefix,
+            return prefix + "builders/%s/builds/%d/steps/%s/logs/%s" % (
                 urllib.quote(builder.getName(), safe=''),
                 build.getNumber(),
-                "step-" + urllib.quote(step.getName(), safe=''),
-                lognum)
+                urllib.quote(step.getName(), safe=''),
+                urllib.quote(log.getName()))
 
     def getChangeSources(self):
         return list(self.botmaster.parent.change_svc)
@@ -1986,10 +2242,11 @@ class Status:
         builder_status.status = self
 
         if not os.path.isdir(builder_status.basedir):
-            os.mkdir(builder_status.basedir)
+            os.makedirs(builder_status.basedir)
         builder_status.determineNextBuildNumber()
 
         builder_status.setBigState("offline")
+        builder_status.setLogCompressionLimit(self.logCompressionLimit)
 
         for t in self.watchers:
             self.announceNewBuilder(t, name, builder_status)
@@ -2000,10 +2257,6 @@ class Status:
         for t in self.watchers:
             t.builderRemoved(name)
 
-    def prune(self):
-        for b in self.botmaster.builders.values():
-            b.builder_status.prune()
-
     def buildsetSubmitted(self, bss):
         self.activeBuildSets.append(bss)
         bss.waitUntilFinished().addCallback(self.activeBuildSets.remove)
