@@ -3,9 +3,14 @@ from __future__ import generators
 import sys, time, os.path
 import urllib
 
+from twisted.web import html, resource
+
 from buildbot import util
 from buildbot import version
 from buildbot.status.web.base import HtmlResource
+#from buildbot.status.web.base import Box, HtmlResource, IBox, ICurrentBox, \
+#     ITopBox, td, build_get_class, path_to_build, path_to_step, map_branches
+from buildbot.status.web.base import build_get_class
 
 # set grid_css to the full pathname of the css file
 if hasattr(sys, "frozen"):
@@ -19,18 +24,7 @@ else:
 
 class ANYBRANCH: pass # a flag value, used below
 
-class GridStatusResource(HtmlResource):
-    # TODO: docs
-    status = None
-    control = None
-    changemaster = None
-
-    def __init__(self, allowForce=True, css=None):
-        HtmlResource.__init__(self)
-
-        self.allowForce = allowForce
-        self.css = css or grid_css
-
+class GridStatusMixin(object):
     def getTitle(self, request):
         status = self.getStatus(request)
         p = status.getProjectName()
@@ -41,7 +35,7 @@ class GridStatusResource(HtmlResource):
 
     def getChangemaster(self, request):
         # TODO: this wants to go away, access it through IStatus
-        return request.site.buildbot_service.parent.change_svc
+        return request.site.buildbot_service.getChangeSvc()
 
     # handle reloads through an http header
     # TODO: send this as a real header, rather than a tag
@@ -79,15 +73,11 @@ class GridStatusResource(HtmlResource):
             return '<td class="build">&nbsp;</td>\n'
 
         if build.isFinished():
-            color = build.getColor()
-            if color == 'green': color = '#72ff75' # the "Buildbot Green"
-
             # get the text and annotate the first line with a link
             text = build.getText()
             if not text: text = [ "(no information)" ]
             if text == [ "build", "successful" ]: text = [ "OK" ]
         else:
-            color = 'yellow' # to match the yellow of the builder
             text = [ 'building' ]
 
         name = build.getBuilder().getName()
@@ -95,8 +85,9 @@ class GridStatusResource(HtmlResource):
         url = "builders/%s/builds/%d" % (name, number)
         text[0] = '<a href="%s">%s</a>' % (url, text[0])
         text = '<br />\n'.join(text)
+        class_ = build_get_class(build)
 
-        return '<td class="build" bgcolor="%s">%s</td>\n' % (color, text)
+        return '<td class="build %s">%s</td>\n' % (class_, text)
 
     def builder_td(self, request, builder):
         state, builds = builder.getState()
@@ -114,17 +105,6 @@ class GridStatusResource(HtmlResource):
         if state == "idle" and upcoming:
             state = "waiting"
 
-        if state == "building":
-            color = "yellow"
-        elif state == "offline":
-            color = "red"
-        elif state == "idle":
-            color = "white"
-        elif state == "waiting":
-            color = "yellow"
-        else:
-            color = "white"
-
         # TODO: for now, this pending/upcoming stuff is in the "current
         # activity" box, but really it should go into a "next activity" row
         # instead. The only times it should show up in "current activity" is
@@ -140,16 +120,64 @@ class GridStatusResource(HtmlResource):
             else:
                 text += "<br />(%s)" % state
 
-        return  '<td valign="center" bgcolor="%s" class="builder">%s</td>\n' % \
-            (color, text)
+        return  '<td valign="center" class="builder %s">%s</td>\n' % \
+            (state, text)
 
     def stamp_td(self, stamp):
         text = stamp.getText()
         return '<td valign="bottom" class="sourcestamp">%s</td>\n' % \
             "<br />".join(text)
 
+    def getRecentSourcestamps(self, status, numBuilds, categories, branch):
+        """
+        get a list of the most recent NUMBUILDS SourceStamp tuples, sorted
+        by the earliest start we've seen for them
+        """
+        # TODO: use baseweb's getLastNBuilds?
+        sourcestamps = { } # { ss-tuple : earliest time }
+        for bn in status.getBuilderNames():
+            builder = status.getBuilder(bn)
+            if categories and builder.category not in categories:
+                continue
+            build = builder.getBuild(-1)
+            while build:
+                ss = build.getSourceStamp(absolute=True)
+                start = build.getTimes()[0]
+                build = build.getPreviousBuild()
+
+                # skip un-started builds
+                if not start: continue
+
+                # skip non-matching branches
+                if branch != ANYBRANCH and ss.branch != branch: continue
+
+                sourcestamps[ss] = min(sourcestamps.get(ss, sys.maxint), start)
+
+        # now sort those and take the NUMBUILDS most recent
+        sourcestamps = sourcestamps.items()
+        sourcestamps.sort(lambda x, y: cmp(x[1], y[1]))
+        sourcestamps = map(lambda tup : tup[0], sourcestamps)
+        sourcestamps = sourcestamps[-numBuilds:]
+
+        return sourcestamps
+
+class GridStatusResource(HtmlResource, GridStatusMixin):
+    # TODO: docs
+    status = None
+    control = None
+    changemaster = None
+
+    def __init__(self, allowForce=True, css=None):
+        HtmlResource.__init__(self)
+
+        self.allowForce = allowForce
+        self.css = css or grid_css
+
+
     def body(self, request):
-        "This method builds the main waterfall display."
+        """This method builds the regular grid display.
+        That is, build stamps across the top, build hosts down the left side
+        """
 
         # get url parameters
         numBuilds = int(request.args.get("width", [5])[0])
@@ -168,12 +196,13 @@ class GridStatusResource(HtmlResource):
         data += '<tr>\n'
         data += '<td class="title"><a href="%s">%s</a>' % (projectURL, projectName)
         if categories:
+            html_categories = map(html.escape(categories))
             if len(categories) > 1:
-                data += '\n<br /><b>Categories:</b><br/>%s' % ('<br/>'.join(categories))
+                data += '\n<br /><b>Categories:</b><br/>%s' % ('<br/>'.join(html_categories))
             else:
-                data += '\n<br /><b>Category:</b> %s' % categories[0]
+                data += '\n<br /><b>Category:</b> %s' % html_categories[0]
         if branch != ANYBRANCH:
-            data += '\n<br /><b>Branch:</b> %s' % (branch or 'trunk')
+            data += '\n<br /><b>Branch:</b> %s' % (html.escape(branch) or 'trunk')
         data += '</td>\n'
         for stamp in stamps:
             data += self.stamp_td(stamp)
@@ -228,36 +257,105 @@ class GridStatusResource(HtmlResource):
         data += '</div>\n'
         return data
 
-    def getRecentSourcestamps(self, status, numBuilds, categories, branch):
+class TransposedGridStatusResource(HtmlResource, GridStatusMixin):
+    # TODO: docs
+    status = None
+    control = None
+    changemaster = None
+
+    def __init__(self, allowForce=True, css=None):
+        HtmlResource.__init__(self)
+
+        self.allowForce = allowForce
+        self.css = css or grid_css
+
+
+    def body(self, request):
+        """This method builds the transposed grid display.
+        That is, build hosts across the top, ebuild stamps down the left side
         """
-        get a list of the most recent NUMBUILDS SourceStamp tuples, sorted
-        by the earliest start we've seen for them
-        """
-        # TODO: use baseweb's getLastNBuilds?
-        sourcestamps = { } # { ss-tuple : earliest time }
-        for bn in status.getBuilderNames():
+
+        # get url parameters
+        numBuilds = int(request.args.get("length", [5])[0])
+        categories = request.args.get("category", [])
+        branch = request.args.get("branch", [ANYBRANCH])[0]
+        if branch == 'trunk': branch = None
+
+        # and the data we want to render
+        status = self.getStatus(request)
+        stamps = self.getRecentSourcestamps(status, numBuilds, categories, branch)
+
+        projectURL = status.getProjectURL()
+        projectName = status.getProjectName()
+
+        data = '<table class="Grid" border="0" cellspacing="0">\n'
+        data += '<tr>\n'
+        data += '<td class="title"><a href="%s">%s</a>' % (projectURL, projectName)
+        if categories:
+            html_categories = map(html.escape(categories))
+            if len(categories) > 1:
+                data += '\n<br /><b>Categories:</b><br/>%s' % ('<br/>'.join(html_categories))
+            else:
+                data += '\n<br /><b>Category:</b> %s' % html_categories[0]
+        if branch != ANYBRANCH:
+            data += '\n<br /><b>Branch:</b> %s' % (html.escape(branch) or 'trunk')
+        data += '</td>\n'
+
+        sortedBuilderNames = status.getBuilderNames()[:]
+        sortedBuilderNames.sort()
+
+        builder_builds = {}
+
+        for bn in sortedBuilderNames:
+            builds = [None] * len(stamps)
+
             builder = status.getBuilder(bn)
             if categories and builder.category not in categories:
                 continue
+
             build = builder.getBuild(-1)
-            while build:
+            while build and None in builds:
                 ss = build.getSourceStamp(absolute=True)
-                start = build.getTimes()[0]
+                for i in range(len(stamps)):
+                    if ss == stamps[i] and builds[i] is None:
+                        builds[i] = build
                 build = build.getPreviousBuild()
 
-                # skip un-started builds
-                if not start: continue
+            data += self.builder_td(request, builder)
+            builder_builds[bn] = builds
 
-                # skip non-matching branches
-                if branch != ANYBRANCH and ss.branch != branch: continue
+        data += '</tr>\n'
 
-                sourcestamps[ss] = min(sourcestamps.get(ss, sys.maxint), start)
+        for i in range(len(stamps)):
+            data += '<tr>\n'
+            data += self.stamp_td(stamps[i])
+            for bn in sortedBuilderNames:
+                data += self.build_td(request, builder_builds[bn][i])
+            data += '</tr>\n'
 
-        # now sort those and take the NUMBUILDS most recent
-        sourcestamps = sourcestamps.items()
-        sourcestamps.sort(lambda x, y: cmp(x[1], y[1]))
-        sourcestamps = map(lambda tup : tup[0], sourcestamps)
-        sourcestamps = sourcestamps[-numBuilds:]
+        data += '</table>\n'
 
-        return sourcestamps
+        # TODO: this stuff should be generated by a template of some sort
+        data += '<hr /><div class="footer">\n'
+
+        welcomeurl = self.path_to_root(request) + "index.html"
+        data += '[<a href="%s">welcome</a>]\n' % welcomeurl
+        data += "<br />\n"
+
+        data += '<a href="http://buildbot.sourceforge.net/">Buildbot</a>'
+        data += "-%s " % version
+        if projectName:
+            data += "working for the "
+            if projectURL:
+                data += "<a href=\"%s\">%s</a> project." % (projectURL,
+                                                            projectName)
+            else:
+                data += "%s project." % projectName
+        data += "<br />\n"
+        data += ("Page built: " +
+                 time.strftime("%a %d %b %Y %H:%M:%S",
+                               time.localtime(util.now()))
+                 + "\n")
+        data += '</div>\n'
+        return data
 
