@@ -269,6 +269,15 @@ appear in-between the actual builds.</p>
 
 %(show_events_input)s
 
+<h2>Showing only the Builders with failures</h2>
+
+<p>By adding the <tt>failures_only=true</tt> argument, the display will be limited
+to showing builders that are currently failing. A builder is considered
+failing if the last finished build was not successful, a step in the current
+build(s) failed, or if the builder is offline.
+
+%(failures_only_input)s
+
 <h2>Showing only Certain Branches</h2>
 
 <p>If you provide one or more <tt>branch=</tt> arguments, the display will be
@@ -326,6 +335,16 @@ class WaterfallHelp(HtmlResource):
                              '</p>\n'
                              ) % showEvents_checked
 
+        failuresOnly_checked = ''
+        if request.args.get("failures_only", ["false"])[0].lower() == "true":
+            failuresOnly_checked = 'checked="checked"'
+        failures_only_input = ('<p>'
+                               '<input type="checkbox" name="failures_only" '
+                               'value="true" %s>'
+                               'Show failures only'
+                               '</p>\n'
+                               ) % failuresOnly_checked
+
         branches = [b
                     for b in request.args.get("branch", [])
                     if b]
@@ -387,6 +406,7 @@ class WaterfallHelp(HtmlResource):
                   "show_branches_input": show_branches_input,
                   "show_builders_input": show_builders_input,
                   "show_reload_input": show_reload_input,
+                  "failures_only_input": failures_only_input,
                   }
         data += HELP % fields
         return data
@@ -428,6 +448,34 @@ class WaterfallStatusResource(HtmlResource):
             head += '<meta http-equiv="refresh" content="%d">\n' % reload_time
         return head
 
+    def isSuccess(self, builderStatus):
+        # Helper function to return True if the builder is not failing.
+        # The function will return false if the current state is "offline",
+        # the last build was not successful, or if a step from the current
+        # build(s) failed.
+
+        # Make sure the builder is online.
+        if builderStatus.getState()[0] == 'offline':
+            return False
+
+        # Look at the last finished build to see if it was success or not.
+        lastBuild = builderStatus.getLastFinishedBuild()
+        if lastBuild and lastBuild.getResults() != builder.SUCCESS:
+            return False
+
+        # Check all the current builds to see if one step is already
+        # failing.
+        currentBuilds = builderStatus.getCurrentBuilds()
+        if currentBuilds:
+            for build in currentBuilds:
+                for step in build.getSteps():
+                    if step.getResults()[0] == builder.FAILURE:
+                        return False
+
+        # The last finished build was successful, and all the current builds
+        # don't have any failed steps.
+        return True
+
     def body(self, request):
         "This method builds the main waterfall display."
 
@@ -462,6 +510,13 @@ class WaterfallStatusResource(HtmlResource):
         if showCategories:
             builders = [b for b in builders if b.category in showCategories]
 
+        # If the URL has the failures_only=true argument, we remove all the
+        # builders that are not currently red or won't be turning red at the end
+        # of their current run.
+        failuresOnly = request.args.get("failures_only", ["false"])[0]
+        if failuresOnly.lower() == "true":
+            builders = [b for b in builders if not self.isSuccess(b)]
+
         builderNames = [b.name for b in builders]
 
         if phase == -1:
@@ -529,7 +584,7 @@ class WaterfallStatusResource(HtmlResource):
                     newargs[k].append(v)
                 else:
                     newargs[k] = [v]
-            newquery = "&".join(["%s=%s" % (k, v)
+            newquery = "&".join(["%s=%s" % (k, urllib.quote(v))
                                  for k in newargs
                                  for v in newargs[k]
                                  ])
@@ -631,6 +686,7 @@ class WaterfallStatusResource(HtmlResource):
         showEvents = False
         if request.args.get("show_events", ["true"])[0].lower() == "true":
             showEvents = True
+        filterCategories = request.args.get('category', [])
         filterBranches = [b for b in request.args.get("branch", []) if b]
         filterBranches = map_branches(filterBranches)
         maxTime = int(request.args.get("last_time", [util.now()])[0])
@@ -677,7 +733,7 @@ class WaterfallStatusResource(HtmlResource):
             return event
 
         for s in sources:
-            gen = insertGaps(s.eventGenerator(filterBranches), lastEventTime)
+            gen = insertGaps(s.eventGenerator(filterBranches, filterCategories), lastEventTime)
             sourceGenerators.append(gen)
             # get the first event
             sourceEvents.append(get_event_from(gen))
@@ -952,7 +1008,11 @@ class WaterfallStatusResource(HtmlResource):
             for strip in grid:
                 b = strip[i]
                 if b:
-                    data += b.td()
+                    # convert data to a unicode string, whacking any non-ASCII characters it might contain
+                    s = b.td()
+                    if isinstance(s, unicode):
+                        s = s.encode("utf-8", "replace")
+                    data += s
                 else:
                     if noBubble:
                         data += td([])
