@@ -18,6 +18,8 @@ from buildbot.status import base
 from buildbot.status.builder import SUCCESS, WARNINGS, FAILURE, EXCEPTION
 from buildbot.scripts.runner import ForceOptions
 
+from string import join, capitalize, lower
+
 class UsageError(ValueError):
     def __init__(self, string = "Invalid usage", *more):
         ValueError.__init__(self, string, *more)
@@ -49,7 +51,7 @@ class IrcBuildRequest:
         self.parent.send(response)
         self.parent.send("I'll give a shout when the build finishes")
         d = s.waitUntilFinished()
-        d.addCallback(self.parent.buildFinished)
+        d.addCallback(self.parent.watchedBuildFinished)
 
 
 class Contact:
@@ -68,6 +70,7 @@ class Contact:
         self.channel = channel
         self.notify_events = {}
         self.subscribed = 0
+	self.add_notification_events(channel.notify_events)
 
     silly = {
         "What happen ?": "Somebody set up us the bomb.",
@@ -168,7 +171,7 @@ class Contact:
     command_STATUS.usage = "status [<which>] - List status of a builder (or all builders)"
 
     def validate_notification_event(self, event):
-        if not re.compile("^(started|finished|success|failed|exception)$").match(event):
+        if not re.compile("^(started|finished|success|failure|exception|warnings|(success|warnings|exception|failure)To(Failure|Success|Warnings|Exception))$").match(event):
             raise UsageError("try 'notify on|off <EVENT>'")
 
     def list_notified_events(self):
@@ -193,14 +196,23 @@ class Contact:
             self.validate_notification_event(event)
             self.notify_events[event] = 1
 
+            if not self.subscribed:
+                self.subscribe_to_build_events()
+
     def remove_notification_events(self, events):
         for event in events:
             self.validate_notification_event(event)
             del self.notify_events[event]
 
+            if len(self.notify_events) == 0 and self.subscribed:
+                self.unsubscribe_from_build_events()
+
     def remove_all_notification_events(self):
         self.notify_events = {}
 
+        if self.subscribed:
+            self.unsubscribe_from_build_events()
+
     def command_NOTIFY(self, args, who):
         args = args.split()
 
@@ -230,13 +242,7 @@ class Contact:
         else:
             raise UsageError("try 'notify on|off <EVENT>'")
 
-        if len(self.notify_events) > 0 and not self.subscribed:
-            self.subscribe_to_build_events()
-
-        elif len(self.notify_events) == 0 and self.subscribed:
-            self.unsubscribe_from_build_events()
-
-    command_NOTIFY.usage = "notify on|off|list [<EVENT>] ... - Notify me about build events.  event should be one or more of: 'started', 'finished', 'failed', 'success', 'exception', 'successToFailed', 'failedToSuccess'"
+    command_NOTIFY.usage = "notify on|off|list [<EVENT>] ... - Notify me about build events.  event should be one or more of: 'started', 'finished', 'failure', 'success', 'exception' or 'xToY' (where x and Y are one of success, warnings, failure, exception, but Y is capitalized)"
 
     def command_WATCH(self, args, who):
         args = args.split()
@@ -271,6 +277,10 @@ class Contact:
     def builderChangedState(self, builderName, state):
         log.msg('[Contact] Builder %s changed state to %s' % (builderName, state))
 
+    def requestSubmitted(self, brstatus):
+        log.msg('[Contact] BuildRequest for %s submitted to Builder %s' %
+            (brstatus.getSourceStamp(), brstatus.builderName))
+
     def builderRemoved(self, builderName):
         log.msg('[Contact] Builder %s removed' % (builderName))
 
@@ -310,9 +320,6 @@ class Contact:
         # only notify about builders we are interested in
         log.msg('[Contact] builder %r in category %s finished' % (builder, builder.category))
 
-        if not self.notify_for('finished', 'failed', 'success', 'exception', 'failedToSuccess', 'successToFailed'):
-            return
-
         if (self.channel.categories != None and
             builder.category not in self.channel.categories):
             return
@@ -328,19 +335,20 @@ class Contact:
         if buildurl:
             r += "  Build details are at %s" % buildurl
 
-        if (self.notify_for('finished')) or \
-           (self.notify_for('success') and results == SUCCESS) or \
-           (self.notify_for('failed') and results == FAILURE) or \
-           (self.notify_for('exception') and results == EXCEPTION):
+        if self.notify_for('finished') or self.notify_for(lower(results_descriptions.get(results))):
             self.send(r)
             return
 
         prevBuild = build.getPreviousBuild()
         if prevBuild:
-            prevResult = prevBuild.getResult()
+            prevResult = prevBuild.getResults()
 
-            if (self.notify_for('failureToSuccess') and prevResult == FAILURE and results == SUCCESS) or \
-               (self.notify_for('successToFailure') and prevResult == SUCCESS and results == FAILURE):
+            required_notification_control_string = join((lower(results_descriptions.get(prevResult)), \
+                                                             'To', \
+                                                             capitalize(results_descriptions.get(results))), \
+                                                            '')
+
+            if (self.notify_for(required_notification_control_string)):
                 self.send(r)
 
     def watchedBuildFinished(self, b):
@@ -377,7 +385,7 @@ class Contact:
             raise UsageError("try 'force build WHICH <REASON>'")
         opts = ForceOptions()
         opts.parseOptions(args)
-        
+
         which = opts['builder']
         branch = opts['branch']
         revision = opts['revision']
@@ -455,8 +463,8 @@ class Contact:
             last = b.getLastFinishedBuild()
             if last:
                 start,finished = last.getTimes()
-                str += ", last build %s secs ago: %s" % \
-                       (int(util.now() - finished), " ".join(last.getText()))
+                str += ", last build %s ago: %s" % \
+                        (self.convertTime(int(util.now() - finished)), " ".join(last.getText()))
         if state == "building":
             t = []
             for build in builds:
@@ -478,7 +486,7 @@ class Contact:
             str = "(no builds run since last restart)"
         else:
             start,finish = last.getTimes()
-            str = "%s secs ago: " % (int(util.now() - finish))
+            str = "%s ago: " % (self.convertTime(int(util.now() - finish)))
             str += " ".join(last.getText())
         self.send("last build [%s]: %s" % (which, str))
 
@@ -580,9 +588,9 @@ class IRCContact(Contact):
     # userJoined(self, user, channel)
 
     def send(self, message):
-        self.channel.msg(self.dest, message)
+        self.channel.msg(self.dest, message.encode("ascii", "replace"))
     def act(self, action):
-        self.channel.me(self.dest, action)
+        self.channel.me(self.dest, action.encode("ascii", "replace"))
 
     def command_JOIN(self, args, who):
         args = args.split()
@@ -654,7 +662,7 @@ class IrcStatusBot(irc.IRCClient):
     """
     implements(IChannel)
 
-    def __init__(self, nickname, password, channels, status, categories):
+    def __init__(self, nickname, password, channels, status, categories, notify_events):
         """
         @type  nickname: string
         @param nickname: the nickname by which this bot should be known
@@ -671,6 +679,7 @@ class IrcStatusBot(irc.IRCClient):
         self.password = password
         self.status = status
         self.categories = categories
+        self.notify_events = notify_events
         self.counter = 0
         self.hasQuit = 0
         self.contacts = {}
@@ -765,13 +774,14 @@ class IrcStatusFactory(ThrottledClientFactory):
     shuttingDown = False
     p = None
 
-    def __init__(self, nickname, password, channels, categories):
+    def __init__(self, nickname, password, channels, categories, notify_events):
         #ThrottledClientFactory.__init__(self) # doesn't exist
         self.status = None
         self.nickname = nickname
         self.password = password
         self.channels = channels
         self.categories = categories
+	self.notify_events = notify_events
 
     def __getstate__(self):
         d = self.__dict__.copy()
@@ -786,7 +796,7 @@ class IrcStatusFactory(ThrottledClientFactory):
     def buildProtocol(self, address):
         p = self.protocol(self.nickname, self.password,
                           self.channels, self.status,
-                          self.categories)
+                          self.categories, self.notify_events)
         p.factory = self
         p.status = self.status
         p.control = self.control
@@ -819,7 +829,7 @@ class IRC(base.StatusReceiverMultiService):
                      "categories"]
 
     def __init__(self, host, nick, channels, port=6667, allowForce=True,
-                 categories=None, password=None):
+                 categories=None, password=None, notify_events={}):
         base.StatusReceiverMultiService.__init__(self)
 
         assert allowForce in (True, False) # TODO: implement others
@@ -832,10 +842,11 @@ class IRC(base.StatusReceiverMultiService):
         self.password = password
         self.allowForce = allowForce
         self.categories = categories
+        self.notify_events = notify_events
 
         # need to stash the factory so we can give it the status object
         self.f = IrcStatusFactory(self.nick, self.password,
-                                  self.channels, self.categories)
+                                  self.channels, self.categories, self.notify_events)
 
         c = internet.TCPClient(host, port, self.f)
         c.setServiceParent(self)
@@ -859,3 +870,4 @@ class IRC(base.StatusReceiverMultiService):
 ## buildbot: status full-2.3
 ##  building, not, % complete, ETA
 ## buildbot: force build full-2.3 "reason"
+
