@@ -1,11 +1,11 @@
 # -*- test-case-name: buildbot.test.test_transfer -*-
 
-import os.path
+import os.path, tarfile, tempfile
 from twisted.internet import reactor
 from twisted.spread import pb
 from twisted.python import log
 from buildbot.process.buildstep import RemoteCommand, BuildStep
-from buildbot.process.buildstep import SUCCESS, FAILURE
+from buildbot.process.buildstep import SUCCESS, FAILURE, SKIPPED
 from buildbot.interfaces import BuildSlaveTooOldError
 
 
@@ -25,16 +25,16 @@ class _FileWriter(pb.Referenceable):
         self.fp = open(destfile, "wb")
         if mode is not None:
             os.chmod(destfile, mode)
-	self.remaining = maxsize
+        self.remaining = maxsize
 
     def remote_write(self, data):
-	"""
-	Called from remote slave to write L{data} to L{fp} within boundaries
-	of L{maxsize}
+        """
+        Called from remote slave to write L{data} to L{fp} within boundaries
+        of L{maxsize}
 
-	@type  data: C{string}
-	@param data: String of data to write
-	"""
+        @type  data: C{string}
+        @param data: String of data to write
+        """
         if self.remaining is not None:
             if len(data) > self.remaining:
                 data = data[:self.remaining]
@@ -59,6 +59,76 @@ class _FileWriter(pb.Referenceable):
             os.unlink(self.destfile)
 
 
+def _extractall(self, path=".", members=None):
+    """Fallback extractall method for TarFile, in case it doesn't have its own."""
+
+    import copy
+    import operator
+
+    directories = []
+
+    if members is None:
+        members = self
+
+    for tarinfo in members:
+        if tarinfo.isdir():
+            # Extract directories with a safe mode.
+            directories.append(tarinfo)
+            tarinfo = copy.copy(tarinfo)
+            tarinfo.mode = 0700
+        self.extract(tarinfo, path)
+
+    # Reverse sort directories.
+    directories.sort(lambda a, b: cmp(a.name, b.name))
+    directories.reverse()
+
+    # Set correct owner, mtime and filemode on directories.
+    for tarinfo in directories:
+        dirpath = os.path.join(path, tarinfo.name)
+        try:
+            self.chown(tarinfo, dirpath)
+            self.utime(tarinfo, dirpath)
+            self.chmod(tarinfo, dirpath)
+        except tarfile.ExtractError, e:
+            if self.errorlevel > 1:
+                raise
+            else:
+                self._dbg(1, "tarfile: %s" % e)
+
+class _DirectoryWriter(_FileWriter):
+    """
+    A DirectoryWriter is implemented as a FileWriter, with an added post-processing
+    step to unpack the archive, once the transfer has completed.
+    """
+
+    def __init__(self, destroot, maxsize, compress, mode):
+        self.destroot = destroot
+
+        self.fd, self.tarname = tempfile.mkstemp()
+        self.compress = compress
+        _FileWriter.__init__(self, self.tarname, maxsize, mode)
+
+    def remote_unpack(self):
+        """
+        Called by remote slave to state that no more data will be transfered
+        """
+        if self.fp:
+            self.fp.close()
+            self.fp = None
+        fileobj = os.fdopen(self.fd, 'r')
+        if self.compress == 'bz2':
+            mode='r|bz2'
+        elif self.compress == 'gz':
+            mode='r|gz'
+        else:
+            mode = 'r'
+        if not hasattr(tarfile.TarFile, 'extractall'):
+            tarfile.TarFile.extractall = _extractall
+        archive = tarfile.open(name=self.tarname, mode=mode, fileobj=fileobj)
+        archive.extractall(path=self.destroot)
+        os.remove(self.tarname)
+
+
 class StatusRemoteCommand(RemoteCommand):
     def __init__(self, remote_command, args):
         RemoteCommand.__init__(self, remote_command, args)
@@ -73,8 +143,40 @@ class StatusRemoteCommand(RemoteCommand):
         if 'stderr' in update:
             self.stderr = self.stderr + update['stderr'] + '\n'
 
+class _TransferBuildStep(BuildStep):
+    """
+    Base class for FileUpload and FileDownload to factor out common
+    functionality.
+    """
+    DEFAULT_WORKDIR = "build"           # is this redundant?
 
-class FileUpload(BuildStep):
+    def setDefaultWorkdir(self, workdir):
+        if self.workdir is None:
+            self.workdir = workdir
+
+    def _getWorkdir(self):
+        properties = self.build.getProperties()
+        if self.workdir is None:
+            workdir = self.DEFAULT_WORKDIR
+        else:
+            workdir = self.workdir
+        return properties.render(workdir)
+
+    def finished(self, result):
+        # Subclasses may choose to skip a transfer. In those cases, self.cmd
+        # will be None, and we should just let BuildStep.finished() handle
+        # the rest
+        if result == SKIPPED:
+            return BuildStep.finished(self, SKIPPED)
+        if self.cmd.stderr != '':
+            self.addCompleteLog('stderr', self.cmd.stderr)
+
+        if self.cmd.rc is None or self.cmd.rc == 0:
+            return BuildStep.finished(self, SUCCESS)
+        return BuildStep.finished(self, FAILURE)
+
+
+class FileUpload(_TransferBuildStep):
     """
     Build step to transfer a file from the slave to the master.
 
@@ -95,7 +197,7 @@ class FileUpload(BuildStep):
     name = 'upload'
 
     def __init__(self, slavesrc, masterdest,
-                 workdir="build", maxsize=None, blocksize=16*1024, mode=None,
+                 workdir=None, maxsize=None, blocksize=16*1024, mode=None,
                  **buildstep_kwargs):
         BuildStep.__init__(self, **buildstep_kwargs)
         self.addFactoryArguments(slavesrc=slavesrc,
@@ -132,7 +234,6 @@ class FileUpload(BuildStep):
         log.msg("FileUpload started, from slave %r to master %r"
                 % (source, masterdest))
 
-        self.step_status.setColor('yellow')
         self.step_status.setText(['uploading', os.path.basename(source)])
 
         # we use maxsize to limit the amount of data on both sides
@@ -141,7 +242,7 @@ class FileUpload(BuildStep):
         # default arguments
         args = {
             'slavesrc': source,
-            'workdir': self.workdir,
+            'workdir': self._getWorkdir(),
             'writer': fileWriter,
             'maxsize': self.maxsize,
             'blocksize': self.blocksize,
@@ -151,20 +252,104 @@ class FileUpload(BuildStep):
         d = self.runCommand(self.cmd)
         d.addCallback(self.finished).addErrback(self.failed)
 
+
+class DirectoryUpload(BuildStep):
+    """
+    Build step to transfer a directory from the slave to the master.
+
+    arguments:
+
+    - ['slavesrc']   name of source directory at slave, relative to workdir
+    - ['masterdest'] name of destination directory at master
+    - ['workdir']    string with slave working directory relative to builder
+                     base dir, default 'build'
+    - ['maxsize']    maximum size of each file, default None (=unlimited)
+    - ['blocksize']  maximum size of each block being transfered
+    - ['compress']   compression type to use: one of [None, 'gz', 'bz2']
+    - ['mode']       file access mode for the resulting master-side file.
+                     The default (=None) is to leave it up to the umask of
+                     the buildmaster process.
+
+    """
+
+    name = 'upload'
+
+    def __init__(self, slavesrc, masterdest,
+                 workdir="build", maxsize=None, blocksize=16*1024, mode=None,
+                 compress=None, **buildstep_kwargs):
+        BuildStep.__init__(self, **buildstep_kwargs)
+        self.addFactoryArguments(slavesrc=slavesrc,
+                                 masterdest=masterdest,
+                                 workdir=workdir,
+                                 maxsize=maxsize,
+                                 blocksize=blocksize,
+                                 compress=compress,
+                                 mode=mode,
+                                 )
+
+        self.slavesrc = slavesrc
+        self.masterdest = masterdest
+        self.workdir = workdir
+        self.maxsize = maxsize
+        self.blocksize = blocksize
+        assert compress in (None, 'gz', 'bz2')
+        self.compress = compress
+        assert isinstance(mode, (int, type(None)))
+        self.mode = mode
+
+    def start(self):
+        version = self.slaveVersion("uploadDirectory")
+        properties = self.build.getProperties()
+
+        if not version:
+            m = "slave is too old, does not know about uploadDirectory"
+            raise BuildSlaveTooOldError(m)
+
+        source = properties.render(self.slavesrc)
+        masterdest = properties.render(self.masterdest)
+        # we rely upon the fact that the buildmaster runs chdir'ed into its
+        # basedir to make sure that relative paths in masterdest are expanded
+        # properly. TODO: maybe pass the master's basedir all the way down
+        # into the BuildStep so we can do this better.
+        masterdest = os.path.expanduser(masterdest)
+        log.msg("DirectoryUpload started, from slave %r to master %r"
+                % (source, masterdest))
+
+        self.step_status.setText(['uploading', os.path.basename(source)])
+        
+        # we use maxsize to limit the amount of data on both sides
+        dirWriter = _DirectoryWriter(masterdest, self.maxsize, self.compress, self.mode)
+
+        # default arguments
+        args = {
+            'slavesrc': source,
+            'workdir': self.workdir,
+            'writer': dirWriter,
+            'maxsize': self.maxsize,
+            'blocksize': self.blocksize,
+            'compress': self.compress
+            }
+
+        self.cmd = StatusRemoteCommand('uploadDirectory', args)
+        d = self.runCommand(self.cmd)
+        d.addCallback(self.finished).addErrback(self.failed)
+
     def finished(self, result):
+        # Subclasses may choose to skip a transfer. In those cases, self.cmd
+        # will be None, and we should just let BuildStep.finished() handle
+        # the rest
+        if result == SKIPPED:
+            return BuildStep.finished(self, SKIPPED)
         if self.cmd.stderr != '':
             self.addCompleteLog('stderr', self.cmd.stderr)
 
         if self.cmd.rc is None or self.cmd.rc == 0:
-            self.step_status.setColor('green')
             return BuildStep.finished(self, SUCCESS)
-        self.step_status.setColor('red')
         return BuildStep.finished(self, FAILURE)
 
 
 
 
-
 class _FileReader(pb.Referenceable):
     """
     Helper class that acts as a file-object with read access
@@ -174,15 +359,15 @@ class _FileReader(pb.Referenceable):
         self.fp = fp
 
     def remote_read(self, maxlength):
-	"""
-	Called from remote slave to read at most L{maxlength} bytes of data
+        """
+        Called from remote slave to read at most L{maxlength} bytes of data
 
-	@type  maxlength: C{integer}
-	@param maxlength: Maximum number of data bytes that can be returned
+        @type  maxlength: C{integer}
+        @param maxlength: Maximum number of data bytes that can be returned
 
         @return: Data read from L{fp}
         @rtype: C{string} of bytes read from file
-	"""
+        """
         if self.fp is None:
             return ''
 
@@ -198,7 +383,7 @@ class _FileReader(pb.Referenceable):
             self.fp = None
 
 
-class FileDownload(BuildStep):
+class FileDownload(_TransferBuildStep):
     """
     Download the first 'maxsize' bytes of a file, from the buildmaster to the
     buildslave. Set the mode of the file
@@ -220,11 +405,10 @@ class FileDownload(BuildStep):
                    the buildslave process.
 
     """
-
     name = 'download'
 
     def __init__(self, mastersrc, slavedest,
-                 workdir="build", maxsize=None, blocksize=16*1024, mode=None,
+                 workdir=None, maxsize=None, blocksize=16*1024, mode=None,
                  **buildstep_kwargs):
         BuildStep.__init__(self, **buildstep_kwargs)
         self.addFactoryArguments(mastersrc=mastersrc,
@@ -258,7 +442,6 @@ class FileDownload(BuildStep):
         log.msg("FileDownload started, from master %r to slave %r" %
                 (source, slavedest))
 
-        self.step_status.setColor('yellow')
         self.step_status.setText(['downloading', "to",
                                   os.path.basename(slavedest)])
 
@@ -281,7 +464,7 @@ class FileDownload(BuildStep):
             'maxsize': self.maxsize,
             'reader': fileReader,
             'blocksize': self.blocksize,
-            'workdir': self.workdir,
+            'workdir': self._getWorkdir(),
             'mode': self.mode,
             }
 
@@ -289,13 +472,3 @@ class FileDownload(BuildStep):
         d = self.runCommand(self.cmd)
         d.addCallback(self.finished).addErrback(self.failed)
 
-    def finished(self, result):
-        if self.cmd.stderr != '':
-            self.addCompleteLog('stderr', self.cmd.stderr)
-
-        if self.cmd.rc is None or self.cmd.rc == 0:
-            self.step_status.setColor('green')
-            return BuildStep.finished(self, SUCCESS)
-        self.step_status.setColor('red')
-        return BuildStep.finished(self, FAILURE)
-
