@@ -15,18 +15,24 @@ from buildbot.changes.changes import Change
 from buildbot.process.builder import Builder
 from time import sleep
 
+import sys
+if sys.version_info[:3] < (2,4,0):
+    from sets import Set as set
+
 mail = None
 try:
     from buildbot.status import mail
 except ImportError:
     pass
 from buildbot.status import progress, client # NEEDS COVERAGE
-from buildbot.test.runutils import RunMixin, setupBuildStepStatus
+from buildbot.test.runutils import RunMixin, setupBuildStepStatus, rmtree
 
 class MyStep:
     build = None
     def getName(self):
         return "step"
+    def getResults(self):
+        return (builder.SUCCESS, "yay")
 
 class MyLogFileProducer(builder.LogFileProducer):
     # The reactor.callLater(0) in LogFileProducer.resumeProducing is a bit of
@@ -134,10 +140,12 @@ def customTextMailMessage(attrs):
     text.append("")
     text.extend([c.asText() for c in attrs['changes']])
     text.append("")
-    name, url, lines = attrs['logs'][-1]
+    name, url, lines, status = attrs['logs'][-1]
     text.append("Last %d lines of '%s':" % (logLines, name))
     text.extend(["\t%s\n" % line for line in lines[len(lines)-logLines:]])
     text.append("")
+    text.append("Build number was: %s" % attrs['buildProperties']['buildnumber'])
+    text.append("")
     text.append("-buildbot")
     return ("\n".join(text), 'plain')
 
@@ -148,11 +156,12 @@ def customHTMLMailMessage(attrs):
                                                           attrs['result'].title()))
     text.append("<h4>Recent Changes:</h4>")
     text.extend([c.asHTML() for c in attrs['changes']])
-    name, url, lines = attrs['logs'][-1]
+    name, url, lines, status = attrs['logs'][-1]
     text.append("<h4>Last %d lines of '%s':</h4>" % (logLines, name))
     text.append("<p>")
     text.append("<br>".join([line for line in lines[len(lines)-logLines:]]))
     text.append("</p>")
+    text.append("<p>Build number was: %s</p>" % attrs['buildProperties']['buildnumber'])
     text.append("<br>")
     text.append("<b>-<a href='%s'>buildbot</a></b>" % attrs['buildbotURL'])
     return ("\n".join(text), 'html')
@@ -317,6 +326,7 @@ class Mail(unittest.TestCase):
         self.messages = []
 
         b1 = self.makeBuild(4, builder.FAILURE)
+        b1.setProperty('buildnumber', 1, 'Build')
         b1.setText(["snarkleack", "polarization", "failed"])
         b1.blamelist = ["dev3", "dev3", "dev3", "dev4",
                         "Thomas_Walters"]
@@ -335,6 +345,7 @@ class Mail(unittest.TestCase):
         self.failUnlessIn("comment1", t)
         self.failUnlessIn("comment2", t)
         self.failUnlessIn("Test 4 failed", t)
+        self.failUnlessIn("number was: 1", t)
 
 
     def testCustomHTMLMessage(self):
@@ -350,6 +361,7 @@ class Mail(unittest.TestCase):
         self.messages = []
 
         b1 = self.makeBuild(4, builder.FAILURE)
+        b1.setProperty('buildnumber', 1, 'Build')
         b1.setText(["snarkleack", "polarization", "failed"])
         b1.blamelist = ["dev3", "dev3", "dev3", "dev4",
                         "Thomas_Walters"]
@@ -368,6 +380,7 @@ class Mail(unittest.TestCase):
         self.failUnlessIn("<h4>Last 3 lines of 'step.test':</h4>", t)
         self.failUnlessIn("<p>Changed by: <b>author2</b><br />", t)
         self.failUnlessIn("Test 3 failed", t)
+        self.failUnlessIn("number was: 1", t)
 
     def testShouldAttachLog(self):
         mailer = mail.MailNotifier(fromaddr="buildbot@example.com", addLogs=True)
@@ -408,6 +421,45 @@ class Mail(unittest.TestCase):
         self.failUnlessEqual(set(r), set(["dev3@dev.com", "dev4@dev.com",
                                  "recip2@example.com", "recip@example.com"]))
 
+    
+    def testChange(self):
+        raise unittest.SkipTest("TODO: Fix Build/Builder mock objects to support getPrevBuild()")
+    
+        mailer = MyMailer(fromaddr="buildbot@example.com", mode="change",
+                          extraRecipients=["bah@bah.bah"],
+                          lookup=MyLookup())
+        mailer.parent = self
+        mailer.status = self
+        self.messages = []
+        
+        b1 = self.makeBuild(1, builder.SUCCESS)
+        b2 = self.makeBuild(2, builder.SUCCESS)
+        b3 = self.makeBuild(3, builder.FAILURE)
+        b4 = self.makeBuild(4, builder.FAILURE)
+        b5 = self.makeBuild(5, builder.SUCCESS)
+        b6 = self.makeBuild(6, builder.SUCCESS)
+        
+        # no message on first or repetetive success
+        mailer.buildFinished("builder1", b1, b1.results)
+        self.failIf(self.messages)
+        mailer.buildFinished("builder1", b2, b2.results)
+        self.failIf(self.messages)
+        
+        # message on first fail only
+        mailer.buildFinished("builder1", b3, b3.results)
+        self.failUnless(len(self.messages) == 1)
+        self.messages.pop()
+        mailer.buildFinished("builder1", b4, b4.results)
+        self.failIf(self.messages)
+
+        # message on first following success
+        mailer.buildFinished("builder1", b5, b5.results)
+        self.failUnless(len(self.messages) == 1)
+        self.messages.pop()
+        mailer.buildFinished("builder1", b6, b6.results)
+        self.failIf(self.messages)
+
+
     def testLogs(self):
         basedir = "test_status_logs"
         os.mkdir(basedir)
@@ -511,10 +563,16 @@ class Results(unittest.TestCase):
         self.failUnlessEqual(t.getLogs(), {'output': ""})
 
 class Log(unittest.TestCase):
-    def setUpClass(self):
-        self.basedir = "status_log_add"
+    basedir = "status_log_add"
+
+    def setUp(self):
+        self.tearDown()
         os.mkdir(self.basedir)
 
+    def tearDown(self):
+        if os.path.exists(self.basedir):
+            rmtree(self.basedir)
+
     def testAdd(self):
         l = MyLog(self.basedir, "compile", step=13)
         self.failUnlessEqual(l.getName(), "compile")
@@ -1629,3 +1687,47 @@ class BuildExpectation(unittest.TestCase):
 
         build.buildFinished(['sometext'], builder.FAILURE)
         self.failUnlessEqual(b.expectations, None, 'Zero expectation for a failed build')
+
+class Pruning(unittest.TestCase):
+    def runTest(self, files, buildHorizon, logHorizon):
+        bstat = builder.BuilderStatus("foo")
+        bstat.buildHorizon = buildHorizon
+        bstat.logHorizon = logHorizon
+        bstat.basedir = "prune-test"
+
+        rmtree(bstat.basedir)
+        os.mkdir(bstat.basedir)
+        for filename in files:
+            open(os.path.join(bstat.basedir, filename), "w").write("TEST")
+        bstat.determineNextBuildNumber()
+
+        bstat.prune()
+
+        remaining = os.listdir(bstat.basedir)
+        remaining.sort()
+        return remaining
+
+    files_base = [
+        '10',
+        '11',
+        '12', '12-log-bar', '12-log-foo',
+        '13', '13-log-foo',
+        '14', '14-log-bar', '14-log-foo',
+    ]
+
+    def test_rmlogs(self):
+        remaining = self.runTest(self.files_base, 5, 2)
+        self.failUnlessEqual(remaining, [
+            '10',
+            '11',
+            '12',
+            '13', '13-log-foo',
+            '14', '14-log-bar', '14-log-foo',
+        ])
+
+    def test_rmbuilds(self):
+        remaining = self.runTest(self.files_base, 2, 0)
+        self.failUnlessEqual(remaining, [
+            '13', '13-log-foo',
+            '14', '14-log-bar', '14-log-foo',
+        ])
